@@ -1,6 +1,7 @@
 #include <meshing.hpp>
 #include <geometry2d.hpp>
 #include <visual.hpp>
+#include <inctcl.hpp>
 
 #include "vsgeom2d.hpp"
 
@@ -20,7 +21,7 @@ namespace netgen
   class SplineGeometryVisRegister : public GeometryRegister
   {
   public:
-    virtual NetgenGeometry * Load (string filename) const { return NULL; }
+    virtual NetgenGeometry * Load (const filesystem::path & filename) const { return NULL; }
     virtual VisualScene * GetVisualScene (const NetgenGeometry * geom) const;
   };
 
@@ -49,6 +50,6 @@ extern "C" int Ng_geom2d_Init (Tcl_Interp * interp);
 
 int Ng_geom2d_Init (Tcl_Interp * interp)
 {
-  geometryregister.Append (new SplineGeometryVisRegister);
+  GeometryRegister().Append(new SplineGeometryVisRegister);
   return TCL_OK;
 }
