@@ -2,12 +2,14 @@
 
 #include <meshing.hpp>
 #include <csg.hpp>
-
+#include <geometry2d.hpp>
 
 #ifdef SOCKETS
 #include "../sockets/sockets.hpp"
 #endif
 
+#include "../general/gzstream.h"
+
 #include "nginterface.h"
 // #include "../visualization/soldata.hpp"
 // #include <visual.hpp>
@@ -17,22 +19,15 @@
 namespace netgen
 {
   DLL_HEADER MeshingParameters mparam;
+
+  /** Force linking of geom2d library (for SplineGeometryRegister)**/
+  SplineGeometry2d dummy_2dgeom; 
 }
 
 static std::thread meshingthread;
 void RunParallel ( void * (*fun)(void *), void * in)
 {
-  bool parthread = netgen::mparam.parthread;
-
-#ifdef PARALLEL
-  int provided;
-  MPI_Query_thread(&provided);
-  if (provided < 3)
-    if (netgen::ntasks > 1) parthread = false;
-  // cout << "runparallel = " << parthread << endl;
-#endif
-
-  if (parthread)
+  if (netgen::mparam.parthread)
     {
       meshingthread = std::thread(fun, in);
       meshingthread.detach();
@@ -55,8 +50,8 @@ namespace netgen
 
 #ifdef SOCKETS
   extern AutoPtr<ClientSocket> clientsocket;
-  //extern Array< AutoPtr < ServerInfo > > servers;
-  extern Array< ServerInfo* > servers;
+  //extern NgArray< AutoPtr < ServerInfo > > servers;
+  extern NgArray< ServerInfo* > servers;
 #endif
 
   
@@ -76,9 +71,9 @@ void Ng_LoadGeometry (const char * filename)
       return;
     }
 
-  for (int i = 0; i < geometryregister.Size(); i++)
+  for (auto loader : GeometryRegister())
     {
-      NetgenGeometry * hgeom = geometryregister[i]->Load (filename);
+      NetgenGeometry * hgeom = loader->Load (filename);
       if (hgeom)
 	{
           ng_geometry.reset (hgeom);
@@ -99,135 +94,168 @@ void Ng_LoadMeshFromStream ( istream & input )
   mesh -> Load(input);
 
   SetGlobalMesh (mesh);
-  for (int i = 0; i < geometryregister.Size(); i++)
-    {
-      NetgenGeometry * hgeom = geometryregister[i]->LoadFromMeshFile (input);
-      if (hgeom)
-	{
-          ng_geometry.reset (hgeom);
-	  break;
-	}
-    }
+  ng_geometry = GeometryRegister().LoadFromMeshFile (input);
+
   if (!ng_geometry)
     ng_geometry = make_shared<NetgenGeometry>();
   mesh->SetGeometry (ng_geometry);
 }
 
 
-
-
-void Ng_LoadMesh (const char * filename)
+void Ng_LoadMesh (const char * filename, ngcore::NgMPI_Comm comm)
 {
+  int id = comm.Rank();
+  int ntasks = comm.Size();
+
   {
-      ifstream infile(filename);
-      if(!infile.good())
-          throw NgException(string("Error opening file ") + filename);
+    ifstream infile(filename);
+    if(!infile.good())
+      throw NgException(string("Error opening file ") + filename);
   }
-#ifdef PARALLEL
-  MPI_Comm_size(MPI_COMM_WORLD, &ntasks);
-  MPI_Comm_rank(MPI_COMM_WORLD, &id);
 
-  if (id == 0)
-    {
-#endif
-      if ( string(filename).find(".vol") == string::npos )
-        /*
-      if ( (strlen (filename) > 4) &&
-	   strcmp (filename + (strlen (filename)-4), ".vol") != 0 )
-        */
-	{
-	  mesh.reset (new Mesh());
-	  ReadFile(*mesh,filename);
-	  
-	  //mesh->SetGlobalH (mparam.maxh);
-	  //mesh->CalcLocalH();
-	  return;
-	}
-
-      string fn(filename);
-
-      istream * infile;
-      if (fn.substr (fn.length()-3, 3) == ".gz")
-        infile = new igzstream (filename);
-      else
-        infile = new ifstream (filename);
-
-      Ng_LoadMeshFromStream(*infile);
-      delete infile;
-      
-#ifdef PARALLEL
-      if (ntasks > 1)
-	{
-
-	  char * weightsfilename = new char [strlen(filename)+1];
-	  strcpy (weightsfilename, filename);            
-	  weightsfilename[strlen (weightsfilename)-3] = 'w';
-	  weightsfilename[strlen (weightsfilename)-2] = 'e';
-	  weightsfilename[strlen (weightsfilename)-1] = 'i';
-
-	  ifstream weightsfile(weightsfilename);      
-	  delete [] weightsfilename;  
-	  
-	  if (!(weightsfile.good()))
-	    {
-	      // cout << "regular distribute" << endl;
-	      mesh -> Distribute();
-	    }
-	  else
-	    {
-	      char str[20];   
-	      bool endfile = false;
-	      int n, dummy;
-	      
-	      Array<int> segment_weights;
-	      Array<int> surface_weights;
-	      Array<int> volume_weights;
-	      
-	      while (weightsfile.good() && !endfile)
-		{
-		  weightsfile >> str;
-		  
-		  if (strcmp (str, "edgeweights") == 0)
-		    {
-		      weightsfile >> n;
-		      segment_weights.SetSize(n);
-		      for (int i = 0; i < n; i++)
-			weightsfile >> dummy >> segment_weights[i];
-		    }
-		  
-		  if (strcmp (str, "surfaceweights") == 0)
-		    {
-		      weightsfile >> n;
-		      surface_weights.SetSize(n);
-		      for (int i=0; i<n; i++)
-			weightsfile >> dummy >> surface_weights[i];
-		    }
-		  
-		  if (strcmp (str, "volumeweights") == 0)
-		    {
-		      weightsfile >> n;
-		      volume_weights.SetSize(n);
-		      for (int i=0; i<n; i++)
-			weightsfile >> dummy >> volume_weights[i];
-		    }
-		  
-		  if (strcmp (str, "endfile") == 0)
-		    endfile = true;  
-		}     
-	      
-	      mesh -> Distribute(volume_weights, surface_weights, segment_weights);
-	    }
-	}
-    }
-  else
+  if ( string(filename).find(".vol") == string::npos )
     {
+      if(ntasks>1)
+	throw NgException("Not sure what to do with this?? Does this work with MPI??");
       mesh.reset (new Mesh());
-//       vssolution.SetMesh(mesh);
-//       vsmesh.SetMesh(mesh);
-      SetGlobalMesh (mesh);
-      mesh->SendRecvMesh();
+      mesh->SetCommunicator(comm);
+      ReadFile(*mesh,filename);
+      //mesh->SetGlobalH (mparam.maxh);
+      //mesh->CalcLocalH();
+      return;
     }
-#endif
+
+  istream * infile;
+  Array<char> buf; // for distributing geometry!
+  int strs;
+
+  if( id == 0) {
+
+    mesh.reset (new Mesh());
+    mesh->SetCommunicator(comm);
+    
+    string fn(filename);
+    if (fn.length() > 8 && fn.substr (fn.length()-8, 8) == ".vol.bin")
+      {
+        mesh -> Load(fn);
+        SetGlobalMesh (mesh);        
+      }
+    else
+      {
+        if (fn.substr (fn.length()-3, 3) == ".gz")
+          infile = new igzstream (filename);
+        else
+          infile = new ifstream (filename);
+        mesh -> Load(*infile);
+        SetGlobalMesh (mesh);
+        
+        // make string from rest of file (for geometry info!)
+        // (this might be empty, in which case we take the global ng_geometry)
+        stringstream geom_part;
+        geom_part << infile->rdbuf();
+        string geom_part_string = geom_part.str();
+        strs = geom_part_string.size();
+        // buf = new char[strs];
+        buf.SetSize(strs);
+        memcpy(buf.Data(), geom_part_string.c_str(), strs*sizeof(char));
+        
+        delete infile;
+      }
+
+    
+    if (ntasks > 1)
+      {
+
+	char * weightsfilename = new char [strlen(filename)+1];
+	strcpy (weightsfilename, filename);            
+	weightsfilename[strlen (weightsfilename)-3] = 'w';
+	weightsfilename[strlen (weightsfilename)-2] = 'e';
+	weightsfilename[strlen (weightsfilename)-1] = 'i';
+
+	ifstream weightsfile(weightsfilename);      
+	delete [] weightsfilename;  
+	  
+	if (!(weightsfile.good()))
+	  {
+	    // cout << "regular distribute" << endl;
+	    mesh -> Distribute();
+	  }
+	else
+	  {
+	    char str[20];   
+	    bool endfile = false;
+	    int n, dummy;
+	      
+	    NgArray<int> segment_weights;
+	    NgArray<int> surface_weights;
+	    NgArray<int> volume_weights;
+	      
+	    while (weightsfile.good() && !endfile)
+	      {
+		weightsfile >> str;
+		  
+		if (strcmp (str, "edgeweights") == 0)
+		  {
+		    weightsfile >> n;
+		    segment_weights.SetSize(n);
+		    for (int i = 0; i < n; i++)
+		      weightsfile >> dummy >> segment_weights[i];
+		  }
+		  
+		if (strcmp (str, "surfaceweights") == 0)
+		  {
+		    weightsfile >> n;
+		    surface_weights.SetSize(n);
+		    for (int i=0; i<n; i++)
+		      weightsfile >> dummy >> surface_weights[i];
+		  }
+		  
+		if (strcmp (str, "volumeweights") == 0)
+		  {
+		    weightsfile >> n;
+		    volume_weights.SetSize(n);
+		    for (int i=0; i<n; i++)
+		      weightsfile >> dummy >> volume_weights[i];
+		  }
+		  
+		if (strcmp (str, "endfile") == 0)
+		  endfile = true;  
+	      }     
+	      
+	    mesh -> Distribute(volume_weights, surface_weights, segment_weights);
+	  }
+      } // ntasks>1 end
+  } // id==0 end
+  else {
+    mesh.reset (new Mesh());
+    mesh->SetCommunicator(comm);
+    SetGlobalMesh (mesh);
+    mesh->SendRecvMesh();
+  }
+
+    /*
+  if(ntasks>1) {
+#ifdef PARALLEL
+    // Scatter the geometry-string (no dummy-implementation in mpi_interface) 
+    int strs = buf.Size();
+    MyMPI_Bcast(strs, comm);
+    if(strs>0)
+      MyMPI_Bcast(buf, comm);
+ #endif
+ }
+  */
+  comm.Bcast(buf);
+
+  shared_ptr<NetgenGeometry> geo;
+  if(buf.Size()) { // if we had geom-info in the file, take it
+    istringstream geom_infile(string((const char*)&buf[0], buf.Size()));
+    geo = GeometryRegister().LoadFromMeshFile(geom_infile);
+  }
+  if(geo!=nullptr) {
+    ng_geometry = geo;
+    mesh->SetGeometry(geo);
+  }
+  else if(ng_geometry!=nullptr) mesh->SetGeometry(ng_geometry);
 }
 
 void Ng_LoadMeshFromString (const char * mesh_as_string)
@@ -449,14 +477,14 @@ const char * Ng_GetDomainMaterial (int dom)
 
 int Ng_GetUserDataSize (char * id)
 {
-  Array<double> da;
+  NgArray<double> da;
   mesh->GetUserData (id, da);
   return da.Size();
 }
 
 void Ng_GetUserData (char * id, double * data)
 {
-  Array<double> da;
+  NgArray<double> da;
   mesh->GetUserData (id, da);
   for (int i = 0; i < da.Size(); i++)
     data[i] = da[i];
@@ -479,7 +507,7 @@ NG_ELEMENT_TYPE Ng_GetSurfaceElement (int ei, int * epi, int * np)
     {
       const Segment & seg = mesh->LineSegment (ei);
 
-      if (seg[2] < 0)
+      if (!seg[2].IsValid())
 	{
 	  epi[0] = seg[0];
 	  epi[1] = seg[1];
@@ -530,7 +558,7 @@ char * Ng_GetSurfaceElementBCName (int ei)
   if ( mesh->GetDimension() == 3 )
     return const_cast<char *>(mesh->GetFaceDescriptor(mesh->SurfaceElement(ei).GetIndex()).GetBCName().c_str());
   else
-    return const_cast<char *>(mesh->LineSegment(ei).GetBCName().c_str());
+    return const_cast<char *>(mesh->GetBCName(mesh->LineSegment(ei).si).c_str());
 }
 
 
@@ -610,12 +638,12 @@ int Ng_FindElementOfPoint (double * p, double * lami, int build_searchtree,
 			   const int * const indices, const int numind)
   
 {
-  Array<int> * dummy(NULL);
+  NgArray<int> * dummy(NULL);
   int ind = -1;
 
   if(indices != NULL)
     {
-      dummy = new Array<int>(numind);
+      dummy = new NgArray<int>(numind);
       for(int i=0; i<numind; i++) (*dummy)[i] = indices[i];
     }
 
@@ -623,14 +651,14 @@ int Ng_FindElementOfPoint (double * p, double * lami, int build_searchtree,
     {
       Point3d p3d(p[0], p[1], p[2]);
       ind = 
-	mesh->GetElementOfPoint(p3d, lami, dummy, build_searchtree != 0);
+	mesh->GetElementOfPoint(p3d, lami, dummy, build_searchtree != 0) + 1;
     }
   else
     {
       double lam3[3];
       Point3d p2d(p[0], p[1], 0);
       ind = 
-	mesh->GetElementOfPoint(p2d, lam3, dummy, build_searchtree != 0);
+	mesh->GetSurfaceElementOfPoint(p2d, lam3, dummy, build_searchtree != 0) + 1;
 
       if (ind > 0)
 	{
@@ -656,12 +684,12 @@ int Ng_FindSurfaceElementOfPoint (double * p, double * lami, int build_searchtre
 				  const int * const indices, const int numind)
   
 {
-  Array<int> * dummy(NULL);
+  NgArray<int> * dummy(NULL);
   int ind = -1;
 
   if(indices != NULL)
     {
-      dummy = new Array<int>(numind);
+      dummy = new NgArray<int>(numind);
       for(int i=0; i<numind; i++) (*dummy)[i] = indices[i];
     }
 
@@ -669,7 +697,7 @@ int Ng_FindSurfaceElementOfPoint (double * p, double * lami, int build_searchtre
     {
       Point3d p3d(p[0], p[1], p[2]);
       ind = 
-	mesh->GetSurfaceElementOfPoint(p3d, lami, dummy, build_searchtree != 0);
+	mesh->GetSurfaceElementOfPoint(p3d, lami, dummy, build_searchtree != 0) + 1;
     }
   else
     {
@@ -838,7 +866,7 @@ NG_ELEMENT_TYPE Ng_GetSegment (int ei, int * epi, int * np)
   epi[0] = seg[0];
   epi[1] = seg[1];
 
-  if (seg[2] < 0)
+  if (!seg[2].IsValid())
     {
       if (np) *np = 2;
       return NG_SEGM;
@@ -875,6 +903,7 @@ void Ng_GetSurfaceElementNeighbouringDomains(const int selnr, int & in, int & ou
 
 // gibt anzahl an distant pnums zurueck
 // * pnums entspricht ARRAY<int[2] >
+[[deprecated("Use GetDistantNodeNums(locnum) -> FlatArray instead!")]]                    
 int NgPar_GetDistantNodeNums ( int nodetype, int locnum, int * distnums )
 {
   int size = NgPar_GetNDistantNodeNums (nodetype, locnum);
@@ -901,6 +930,7 @@ int NgPar_GetDistantNodeNums ( int nodetype, int locnum, int * distnums )
   return size;
 }
 
+[[deprecated("Use GetDistantNodeNums(locnum) -> FlatArray instead!")]]                    
 int NgPar_GetNDistantNodeNums ( int nodetype, int locnum )
 {
   locnum++;
@@ -914,6 +944,7 @@ int NgPar_GetNDistantNodeNums ( int nodetype, int locnum )
   return -1;
 }
 
+[[deprecated("Use GetDistantNodeNums(locnum) -> FlatArray instead!")]]                    
 int NgPar_GetGlobalNodeNum (int nodetype, int locnum)
 {
   locnum++;
@@ -1081,7 +1112,7 @@ void Ng_HPRefinement (int levels, double parameter, bool setorders,
 {
   NgLock meshlock (mesh->MajorMutex(), true);
   Refinement & ref = const_cast<Refinement&> (mesh->GetGeometry()->GetRefinement());
-  HPRefinement (*mesh, &ref, levels, parameter, setorders, ref_level);
+  HPRefinement (*mesh, &ref, SPLIT_HP, levels, parameter, setorders, ref_level);
   /*
     Refinement * ref;
 
@@ -1520,7 +1551,7 @@ int Ng_GetSurfaceElement_Edges (int elnr, int * edges, int * orient)
   /*
     int i, ned;
     const MeshTopology & topology = mesh->GetTopology();
-    Array<int> ia;
+    NgArray<int> ia;
     topology.GetSurfaceElementEdges (elnr, ia);
     ned = ia.Size();
     for (i = 1; i <= ned; i++)
@@ -1540,10 +1571,11 @@ int Ng_GetSurfaceElement_Face (int selnr, int * orient)
 {
   if (mesh->GetDimension() == 3)
     {
+      SurfaceElementIndex sei = selnr-1;
       const MeshTopology & topology = mesh->GetTopology();
       if (orient)
 	*orient = topology.GetSurfaceElementFaceOrientation (selnr);
-      return topology.GetSurfaceElementFace (selnr);
+      return topology.GetFace(sei);
     }
   return -1;
 }
@@ -1551,7 +1583,7 @@ int Ng_GetSurfaceElement_Face (int selnr, int * orient)
 int Ng_GetFace_Vertices (int fnr, int * vert)
 {
   const MeshTopology & topology = mesh->GetTopology();
-  ArrayMem<int,4> ia;
+  NgArrayMem<int,4> ia;
   topology.GetFaceVertices (fnr, ia);
   for (int i = 0; i < ia.Size(); i++)
     vert[i] = ia[i];
@@ -1563,7 +1595,7 @@ int Ng_GetFace_Vertices (int fnr, int * vert)
 int Ng_GetFace_Edges (int fnr, int * edge)
 {
   const MeshTopology & topology = mesh->GetTopology();
-  ArrayMem<int,4> ia;
+  NgArrayMem<int,4> ia;
   topology.GetFaceEdges (fnr, ia);
   for (int i = 0; i < ia.Size(); i++)
     edge[i] = ia[i];
@@ -1573,7 +1605,11 @@ int Ng_GetFace_Edges (int fnr, int * edge)
 void Ng_GetEdge_Vertices (int ednr, int * vert)
 {
   const MeshTopology & topology = mesh->GetTopology();
-  topology.GetEdgeVertices (ednr, vert[0], vert[1]);
+  // topology.GetEdgeVertices (ednr, vert[0], vert[1]);
+  // tie(vert[0], vert[1]) = topology.GetEdgeVertices(ednr-1);
+  auto [v1,v2] = topology.GetEdgeVertices(ednr-1);
+  vert[0] = v1-IndexBASE<PointIndex>()+1;
+  vert[1] = v2-IndexBASE<PointIndex>()+1;
 }
 
 
@@ -1608,19 +1644,19 @@ void Ng_GetVertexElements (int vnr, int * els)
     {
     case 3:
       {
-        FlatArray<ElementIndex> ia = mesh->GetTopology().GetVertexElements(vnr);
+        auto ia = mesh->GetTopology().GetVertexElements(vnr);
         for (int i = 0; i < ia.Size(); i++) els[i] = ia[i]+1;
         break;
       }
     case 2:
       {
-        FlatArray<SurfaceElementIndex> ia = mesh->GetTopology().GetVertexSurfaceElements(vnr);
+        auto ia = mesh->GetTopology().GetVertexSurfaceElements(vnr);
         for (int i = 0; i < ia.Size(); i++) els[i] = ia[i]+1;
         break;
       }
     case 1:
       {
-        FlatArray<SegmentIndex> ia = mesh->GetTopology().GetVertexSegments(vnr);
+        auto ia = mesh->GetTopology().GetVertexSegments(vnr);
         for (int i = 0; i < ia.Size(); i++) els[i] = ia[i]+1;
         break;
         /*
@@ -1694,7 +1730,9 @@ void Ng_SetSurfaceElementOrders (int enr, int ox, int oy)
 
 int Ng_GetNLevels ()
 {
-  return (mesh) ? mesh->mglevels : 0;
+  if (!mesh) return 0;
+  return max(size_t(1), mesh -> level_nv.Size());  
+  // return (mesh) ? mesh->mglevels : 0;
 }
 
 
@@ -1702,8 +1740,8 @@ void Ng_GetParentNodes (int ni, int * parents)
 {
   if (ni <= mesh->mlbetweennodes.Size())
     {
-      parents[0] = mesh->mlbetweennodes.Get(ni).I1();
-      parents[1] = mesh->mlbetweennodes.Get(ni).I2();
+      parents[0] = mesh->mlbetweennodes[ni].I1();
+      parents[1] = mesh->mlbetweennodes[ni].I2();
     }
   else
     parents[0] = parents[1] = 0;
@@ -1715,12 +1753,12 @@ int Ng_GetParentElement (int ei)
   if (mesh->GetDimension() == 3)
     {
       if (ei <= mesh->mlparentelement.Size())
-	return mesh->mlparentelement.Get(ei);
+	return mesh->mlparentelement[ei-1]+1;
     }
   else
     {
       if (ei <= mesh->mlparentsurfaceelement.Size())
-	return mesh->mlparentsurfaceelement.Get(ei);
+	return mesh->mlparentsurfaceelement[ei-1]+1;
     }
   return 0;
 }
@@ -1731,7 +1769,7 @@ int Ng_GetParentSElement (int ei)
   if (mesh->GetDimension() == 3)
     {
       if (ei <= mesh->mlparentsurfaceelement.Size())
-	return mesh->mlparentsurfaceelement.Get(ei);
+	return mesh->mlparentsurfaceelement[ei-1]+1;
     }
   else
     {
@@ -1770,7 +1808,7 @@ int Ng_GetClusterRepElement (int pi)
 		
 int Ng_GetNPeriodicVertices (int idnr)
 {
-  Array<INDEX_2> apairs;
+  NgArray<INDEX_2> apairs;
   mesh->GetIdentifications().GetPairs (idnr, apairs);
   return apairs.Size();
 }
@@ -1779,7 +1817,7 @@ int Ng_GetNPeriodicVertices (int idnr)
 // pairs should be an integer array of 2*npairs
 void Ng_GetPeriodicVertices (int idnr, int * pairs)
 {
-  Array<INDEX_2> apairs;
+  NgArray<INDEX_2> apairs;
   mesh->GetIdentifications().GetPairs (idnr, apairs);
   for (int i = 0; i < apairs.Size(); i++)
     {
@@ -1793,7 +1831,7 @@ void Ng_GetPeriodicVertices (int idnr, int * pairs)
 
 int Ng_GetNPeriodicEdges (int idnr)
 {
-  Array<int,PointIndex::BASE> map;
+  idmap_type map;
   //const MeshTopology & top = mesh->GetTopology();
   int nse = mesh->GetNSeg();
 
@@ -1820,7 +1858,7 @@ int Ng_GetNPeriodicEdges (int idnr)
 
 void Ng_GetPeriodicEdges (int idnr, int * pairs)
 {
-  Array<int,PointIndex::BASE> map;
+  idmap_type map;
   const MeshTopology & top = mesh->GetTopology();
   int nse = mesh->GetNSeg();
 
@@ -1838,8 +1876,10 @@ void Ng_GetPeriodicEdges (int idnr, int * pairs)
 	if (other1 && other2 && mesh->IsSegment (other1, other2))
 	  {
 	    SegmentIndex otherseg = mesh->SegmentNr (other1, other2);
-	    pairs[cnt++] = top.GetSegmentEdge (si+1);
-	    pairs[cnt++] = top.GetSegmentEdge (otherseg+1);
+	    // pairs[cnt++] = top.GetSegmentEdge (si+1);
+	    // pairs[cnt++] = top.GetSegmentEdge (otherseg+1);
+	    pairs[cnt++] = top.GetEdge (si)+1;
+	    pairs[cnt++] = top.GetEdge (otherseg)+1;
 	  }
       }
   }
@@ -1898,8 +1938,9 @@ int Ng_IsRunning()
 int Ng_GetVertex_Elements( int vnr, int* elems )
 {
   const MeshTopology& topology = mesh->GetTopology();
-  ArrayMem<ElementIndex,4> indexArray;
-  topology.GetVertexElements( vnr, indexArray );
+  // ArrayMem<ElementIndex,4> indexArray;
+  // topology.GetVertexElements( vnr, indexArray );
+  auto indexArray = topology.GetVertexElements( vnr );
   
   for( int i=0; i<indexArray.Size(); i++ )
     elems[i] = indexArray[i]+1;
@@ -1908,15 +1949,18 @@ int Ng_GetVertex_Elements( int vnr, int* elems )
 }
 
 ///// Added by Roman Stainko ....
-int Ng_GetVertex_SurfaceElements( int vnr, int* elems )
+int Ng_GetVertex_SurfaceElements( int vnr_, int* elems )
 {
+  PointIndex vnr = vnr_ + IndexBASE<PointIndex>()-1;
+  
   switch (mesh->GetDimension())
     {
     case 3:
       {
         const MeshTopology& topology = mesh->GetTopology();
-        ArrayMem<SurfaceElementIndex,4> indexArray;
-        topology.GetVertexSurfaceElements( vnr, indexArray );
+        // ArrayMem<SurfaceElementIndex,4> indexArray;
+        // topology.GetVertexSurfaceElements( vnr, indexArray );
+        auto indexArray = topology.GetVertexSurfaceElements( vnr );
         
         for( int i=0; i<indexArray.Size(); i++ )
           elems[i] = indexArray[i]+1;
@@ -1947,22 +1991,26 @@ int Ng_GetVertex_SurfaceElements( int vnr, int* elems )
 int Ng_GetVertex_NElements( int vnr )
 {
   const MeshTopology& topology = mesh->GetTopology();
+  /*
   ArrayMem<ElementIndex,4> indexArray;
   topology.GetVertexElements( vnr, indexArray );
-  
   return indexArray.Size();
+  */
+  return topology.GetVertexElements(vnr).Size();
 }
 
 ///// Added by Roman Stainko ....
-int Ng_GetVertex_NSurfaceElements( int vnr )
+int Ng_GetVertex_NSurfaceElements( int vnr_ )
 {
+  PointIndex vnr = vnr_ + IndexBASE<PointIndex>()-1;
   switch (mesh->GetDimension())
     {
     case 3:
       {
         const MeshTopology& topology = mesh->GetTopology();
-        ArrayMem<SurfaceElementIndex,4> indexArray;
-        topology.GetVertexSurfaceElements( vnr, indexArray );
+        // ArrayMem<SurfaceElementIndex,4> indexArray;
+        // topology.GetVertexSurfaceElements( vnr, indexArray );
+        auto indexArray = topology.GetVertexSurfaceElements( vnr );
         return indexArray.Size();
       }
     case 2:
@@ -2033,7 +2081,7 @@ void Ng_SocketClientGetServerClientID ( const int number, int * id )
 
 
 
-
+/*
 #ifdef PARALLEL
 void Ng_SetElementPartition ( const int elnr, const int part )
 {
@@ -2045,6 +2093,7 @@ int Ng_GetElementPartition ( const int elnr )
   return mesh->VolumeElement(elnr+1).GetPartition();
 }
 #endif
+*/
 
 
 void Ng_InitPointCurve(double red, double green, double blue)
@@ -2126,9 +2175,9 @@ int Ng_Bisect_WithInfo ( const char * refinementfile, double ** qualityloss, int
   
   mesh->LocalHFunction().SetGrading (mparam.grading);
 
-  Array<double> * qualityloss_arr = NULL;
+  NgArray<double> * qualityloss_arr = NULL;
   if(qualityloss != NULL)
-    qualityloss_arr = new Array<double>;
+    qualityloss_arr = new NgArray<double>;
 
   ref -> Bisect (*mesh, biopt, qualityloss_arr);
 
@@ -2198,16 +2247,14 @@ int Ng_GetClosureNodes (int nt, int nodenr, int nodeset, int * nodes)
             for (int i = 0; i < el.GetNP(); i++)
               { 
                 nodes[cnt++] = 0;
-                nodes[cnt++] = el[i] - PointIndex::BASE;
+                nodes[cnt++] = el[i] - IndexBASE<PointIndex>();
               }
           }
 
         if (nodeset & 2)  // Edges
           {
-            int edges[12];
-            int ned;
-            ned = mesh->GetTopology().GetElementEdges (nodenr+1, edges, 0);
-            for (int i = 0; i < ned; i++)
+            auto edges = mesh->GetTopology().GetEdges (ElementIndex(nodenr));
+            for (int i = 0; i < edges.Size(); i++)
               {
                 nodes[cnt++] = 1;
                 nodes[cnt++] = edges[i]-1;
@@ -2282,7 +2329,7 @@ int Ng_GetElementClosureNodes (int dim, int elementnr, int nodeset, int * nodes)
             for (int i = 0; i < el.GetNP(); i++)
               { 
                 nodes[cnt++] = 0;
-                nodes[cnt++] = el[i] - PointIndex::BASE;
+                nodes[cnt++] = el[i] - IndexBASE<PointIndex>();
               }
           }
 
@@ -2300,7 +2347,7 @@ int Ng_GetElementClosureNodes (int dim, int elementnr, int nodeset, int * nodes)
 
         if (nodeset & 4)  // Faces
           {
-            int face = mesh->GetTopology().GetSurfaceElementFace (elementnr+1);
+            int face = mesh->GetTopology().GetFace (SurfaceElementIndex(elementnr))+1;
             nodes[cnt++] = 2;
             nodes[cnt++] = face-1;
           }
@@ -2325,13 +2372,6 @@ void Ng_GetArgs (int & argc, char ** &argv)
 
 
 
-void LinkFunction ()
-{
-  Ng_Redraw();
-}
-
-
-
 void Ng_TclCmd(string cmd)
 {
   lock_guard<mutex> guard(tcl_todo_mutex);
