@@ -29,7 +29,7 @@ struct GrowthVectorLimiter
   Array<PointIndex, PointIndex> map_from;
   Table<SurfaceElementIndex, PointIndex> p2sel;
 
-  GrowthVectorLimiter(BoundaryLayerTool& tool_)
+  GrowthVectorLimiter (BoundaryLayerTool& tool_)
     : tool(tool_), params(tool_.params), mesh(tool_.mesh), height(tool_.total_height), growthvectors(tool_.growthvectors), map_from(mesh.Points().Size())
   {
     changed_domains = tool.domains;
@@ -287,7 +287,7 @@ struct GrowthVectorLimiter
     if (factor == 0.0)
       return;
     // for (PointIndex pi : IntRange(tool.np, mesh.GetNP()))
-    for (PointIndex pi : mesh.Points().Range().Modify(tool.np,0))
+    for (PointIndex pi : mesh.Points().Range().Modify(tool.np, 0))
       {
         // auto pi_from = map_from[pi];
         std::set<PointIndex> pis;
@@ -464,8 +464,7 @@ struct GrowthVectorLimiter
   }
 
   template <typename TFunc>
-  void FindTreeIntersections (double trig_shift, double seg_shift, TFunc f,
-                              TBitArray<PointIndex> * relevant_points = nullptr)
+  void FindTreeIntersections (double trig_shift, double seg_shift, TFunc f, TBitArray<PointIndex>* relevant_points = nullptr)
   {
     static Timer t("GrowthVectorLimiter::FindTreeIntersections");
     RegionTimer rt(t);
@@ -506,6 +505,25 @@ struct GrowthVectorLimiter
     RegionTimer reg(t);
     // check if surface trigs are intersecting each other
     bool changed = true;
+    std::set<PointIndex> special_points;
+
+    if (tool.insert_only_volume_elements)
+      for (auto [pi, special_point] : tool.special_boundary_points)
+        {
+          special_points.insert(pi);
+          for (auto& group : special_point.growth_groups)
+            special_points.insert(group.new_points.Last());
+        }
+
+    auto skip_trig = [&] (const Element2d& tri) {
+      if (!tool.insert_only_volume_elements)
+        return false;
+      for (auto pi : tri.PNums())
+        if (special_points.find(pi) != special_points.end())
+          return true;
+      return false;
+    };
+
     while (changed)
       {
         changed = false;
@@ -517,6 +535,9 @@ struct GrowthVectorLimiter
           {
             const Element2d& tri = Get(sei);
 
+            if (skip_trig(tri))
+              continue;
+
             Box<3> box(Box<3>::EMPTY_BOX);
             for (PointIndex pi : tri.PNums())
               box.Add(GetPoint(pi, 1.0, true));
@@ -529,6 +550,9 @@ struct GrowthVectorLimiter
           {
             const Element2d& tri = Get(sei);
 
+            if (skip_trig(tri))
+              continue;
+
             Box<3> box(Box<3>::EMPTY_BOX);
             for (PointIndex pi : tri.PNums())
               box.Add(GetPoint(pi, 1.0, true));
@@ -629,7 +653,7 @@ struct GrowthVectorLimiter
     size_t limit_counter = 1;
 
     TBitArray<PointIndex> relevant_points, relevant_points_next;
-    relevant_points.SetSize(mesh.Points().Size() + 1);   
+    relevant_points.SetSize(mesh.Points().Size() + 1);
     relevant_points_next.SetSize(mesh.Points().Size() + 1);
     relevant_points.Set();
 
@@ -685,8 +709,9 @@ struct GrowthVectorLimiter
     for (auto pi : Range(growthvectors))
       check_point(pi);
 
-    for (auto& [special_pi, special_point] : tool.special_boundary_points)
-      check_point(special_pi);
+    if (!tool.insert_only_volume_elements)
+      for (auto& [special_pi, special_point] : tool.special_boundary_points)
+        check_point(special_pi);
   }
 
   void Perform ()
