@@ -668,7 +668,7 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
                   [](const FaceDescriptor& self)
                   {
                     auto sc = self.SurfColour();
-                    return py::make_tuple(sc[0], sc[1], sc[2]);
+                    return py::make_tuple(sc[0], sc[1], sc[2], sc[3]);
                   },
                   [](FaceDescriptor& self, py::tuple col)
                   {
@@ -694,8 +694,14 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
                   })
     ;
 
+  py::implicitly_convertible< int, SurfaceElementIndex>();
+  PYBIND11_NUMPY_DTYPE(SurfaceElementIndex, i);
+  ExportArray<SurfaceElementIndex, SurfaceElementIndex>(m);
+  
+  py::implicitly_convertible< int, ElementIndex>();
+  PYBIND11_NUMPY_DTYPE(ElementIndex, i);
+  ExportArray<ElementIndex, ElementIndex>(m);
   
-
   ExportArray<Element,ElementIndex>(m);
   ExportArray<Element2d,SurfaceElementIndex>(m);
   ExportArray<Segment,SegmentIndex>(m);
@@ -915,7 +921,7 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
 	    shared_ptr<NetgenGeometry> geo;
 	    if(buf.Size()) { // if we had geom-info in the file, take it
 	      istringstream geom_infile(string((const char*)buf.Data(), buf.Size()));
-	      geo = geometryregister.LoadFromMeshFile(geom_infile);
+	      geo = GeometryRegister().LoadFromMeshFile(geom_infile);
 	    }
 	    if(geo!=nullptr) mesh->SetGeometry(geo);
 	    else if(ng_geometry!=nullptr) mesh->SetGeometry(ng_geometry);
@@ -1300,6 +1306,10 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
     .def("IdentifyPeriodicBoundaries", &Mesh::IdentifyPeriodicBoundaries,
          py::arg("identification_name"), py::arg("face1"), py::arg("mapping"),
 py::arg("point_tolerance") = -1.)
+    .def("GetCurveOrder", [] (Mesh & self)
+          {
+            return self.GetCurvedElements().GetOrder();
+          })
     .def("GetNrIdentifications", [](Mesh& self)
                                  {
                                    return self.GetIdentifications().GetMaxNr();
@@ -1480,7 +1490,8 @@ py::arg("point_tolerance") = -1.)
            }))
     */
     
-    .def ("BuildSearchTree", &Mesh::BuildElementSearchTree,py::call_guard<py::gil_scoped_release>())
+    .def ("BuildSearchTree", &Mesh::BuildElementSearchTree,py::call_guard<py::gil_scoped_release>(),
+          py::arg("dim")=3)
 
     .def ("BoundaryLayer2", GenerateBoundaryLayer2, py::arg("domain"), py::arg("thicknesses"), py::arg("make_new_domain")=true, py::arg("boundaries")=Array<int>{})
     .def ("BoundaryLayer", [](Mesh & self, variant<string, int, std::vector<int>> boundary,
@@ -1682,25 +1693,25 @@ py::arg("point_tolerance") = -1.)
                     return mp;
                   }), py::arg("mp")=nullptr, meshingparameter_description.c_str())
     .def("__str__", &ToString<MP>)
-    .def("RestrictH", [](MP & mp, double x, double y, double z, double h)
+    .def("RestrictH", [](MP & mp, double x, double y, double z, double h, int layer)
           {
-            mp.meshsize_points.Append ( MeshingParameters::MeshSizePoint(Point<3> (x,y,z), h));
-          }, py::arg("x"), py::arg("y"), py::arg("z"), py::arg("h")
+            mp.meshsize_points.Append ( MeshingParameters::MeshSizePoint(Point<3> (x,y,z), h, layer));
+          }, py::arg("x"), py::arg("y"), py::arg("z"), py::arg("h"), py::arg("layer")=1
          )
-    .def("RestrictH", [](MP & mp, const Point<3>& p, double h)
+    .def("RestrictH", [](MP & mp, const Point<3>& p, double h, int layer)
     {
-      mp.meshsize_points.Append ({p, h});
-    }, py::arg("p"), py::arg("h"))
+      mp.meshsize_points.Append ({p, h, layer});
+    }, py::arg("p"), py::arg("h"), py::arg("layer")=1)
     .def("RestrictHLine", [](MP& mp, const Point<3>& p1, const Point<3>& p2,
-                             double maxh)
+                             double maxh, int layer)
     {
       int steps = int(Dist(p1, p2) / maxh) + 2;
       auto v = p2 - p1;
       for (int i = 0; i <= steps; i++)
         {
-          mp.meshsize_points.Append({p1 + double(i)/steps * v, maxh});
+          mp.meshsize_points.Append({p1 + double(i)/steps * v, maxh, layer});
         }
-    }, py::arg("p1"), py::arg("p2"), py::arg("maxh"))
+    }, py::arg("p1"), py::arg("p2"), py::arg("maxh"), py::arg("layer")=1)
     ;
 
   m.def("SetTestoutFile", FunctionPointer ([] (const string & filename)
